@@ -25,13 +25,13 @@
  * list of people on the gedit Team.  
  * See the ChangeLog files for a list of changes. 
  *
- * $Id: gedit-encodings-dialog.h 4429 2005-12-12 17:28:04Z pborelli $
+ * $Id: gedit-encodings-dialog.h 6631 2008-11-29 11:06:49Z pborelli $
  */
 
 #ifndef __GEDIT_ENCODINGS_DIALOG_H__
 #define __GEDIT_ENCODINGS_DIALOG_H__
 
-#include <gtk/gtkdialog.h>
+#include <gtk/gtk.h>
 
 G_BEGIN_DECLS
 
