@@ -81,6 +81,13 @@ struct _AsyncNode
 	GSList *original_children;
 };
 
+typedef struct {
+	GeditFileBrowserStore * model;
+	gchar * virtual_root;
+	GMountOperation * operation;
+	GCancellable * cancellable;
+} MountInfo;
+
 struct _FileBrowserNode 
 {
 	GFile *file;
@@ -119,6 +126,7 @@ struct _GeditFileBrowserStorePrivate
 	SortFunc sort_func;
 
 	GSList *async_handles;
+	MountInfo *mount_info;
 };
 
 static FileBrowserNode *model_find_node 		    (GeditFileBrowserStore *model,
@@ -223,6 +231,17 @@ enum
 
 static guint model_signals[NUM_SIGNALS] = { 0 };
 
+static void
+cancel_mount_operation (GeditFileBrowserStore *obj)
+{
+	if (obj->priv->mount_info != NULL)
+	{
+		obj->priv->mount_info->model = NULL;
+		g_cancellable_cancel (obj->priv->mount_info->cancellable);
+		obj->priv->mount_info = NULL;
+	}
+}
+
 static void
 gedit_file_browser_store_finalize (GObject * object)
 {
@@ -240,6 +259,8 @@ gedit_file_browser_store_finalize (GObject * object)
 		
 		data->removed = TRUE;
 	}
+	
+	cancel_mount_operation (obj);
 
 	g_slist_free (obj->priv->async_handles);
 	G_OBJECT_CLASS (gedit_file_browser_store_parent_class)->finalize (object);
@@ -609,7 +630,7 @@ gedit_file_browser_store_get_path_real (GeditFileBrowserStore * model,
 
 			check = (FileBrowserNode *) (item->data);
 
-			if (model_node_visibility (model, check)) {
+			if (model_node_visibility (model, check) && (check == node || check->inserted)) {
 				if (check == node) {
 					gtk_tree_path_prepend_index (path,
 								     num);
@@ -1267,7 +1288,7 @@ file_browser_node_init (FileBrowserNode * node, GFile * file,
 static FileBrowserNode *
 file_browser_node_new (GFile * file, FileBrowserNode * parent)
 {
-	FileBrowserNode *node = g_new0 (FileBrowserNode, 1);
+	FileBrowserNode *node = g_slice_new0 (FileBrowserNode);
 
 	file_browser_node_init (node, file, parent);
 	return node;
@@ -1278,7 +1299,7 @@ file_browser_node_dir_new (GeditFileBrowserStore * model,
 			   GFile * file, FileBrowserNode * parent)
 {
 	FileBrowserNode *node =
-	    (FileBrowserNode *) g_new0 (FileBrowserNodeDir, 1);
+	    (FileBrowserNode *) g_slice_new0 (FileBrowserNodeDir);
 
 	file_browser_node_init (node, file, parent);
 
@@ -1360,7 +1381,11 @@ file_browser_node_free (GeditFileBrowserStore * model,
 		g_object_unref (node->emblem);
 
 	g_free (node->name);
-	g_free (node);
+	
+	if (NODE_IS_DIR (node))
+		g_slice_free (FileBrowserNodeDir, (FileBrowserNodeDir *)node);
+	else
+		g_slice_free (FileBrowserNode, (FileBrowserNode *)node);
 }
 
 /**
@@ -1401,7 +1426,7 @@ model_remove_node_children (GeditFileBrowserStore * model,
 		
 		return;
 	}
-
+	
 	if (path == NULL)
 		path_child =
 		    gedit_file_browser_store_get_path_real (model, node);
@@ -1706,7 +1731,7 @@ model_check_dummy (GeditFileBrowserStore * model, FileBrowserNode * node)
 				path =
 				    gedit_file_browser_store_get_path_real
 				    (model, dummy);
-
+				    
 				row_inserted (model, &path, &iter);
 				gtk_tree_path_free (path);
 			}
@@ -2647,14 +2672,6 @@ model_root_mounted (GeditFileBrowserStore * model, gchar const * virtual_root)
 	return GEDIT_FILE_BROWSER_STORE_RESULT_OK;
 }
 
-typedef struct {
-	GeditFileBrowserStore * model;
-	gchar * virtual_root;
-	GMountOperation * operation;
-	GCancellable * cancellable;
-} MountInfo;
-
-
 static void
 handle_root_error (GeditFileBrowserStore * model, GError *error)
 {
@@ -2690,12 +2707,15 @@ mount_cb (GFile * file,
 	GeditFileBrowserStore * model = mount_info->model;
 	
 	mounted = g_file_mount_enclosing_volume_finish (file, res, &error);
-	
-	if (g_cancellable_is_cancelled (mount_info->cancellable))
+
+	if (mount_info->model)
+	{
+		model->priv->mount_info = NULL;	
+		model_end_loading (model, model->priv->root);
+	}
+
+	if (!mount_info->model || g_cancellable_is_cancelled (mount_info->cancellable))
 	{
-		if (error)
-			g_error_free (error);
-		
 		// Reset because it might be reused?
 		g_cancellable_reset (mount_info->cancellable);
 	}
@@ -2744,12 +2764,16 @@ model_mount_root (GeditFileBrowserStore * model, gchar const * virtual_root)
 			mount_info->operation = gtk_mount_operation_new (NULL);
 			mount_info->cancellable = g_object_ref (FILE_BROWSER_NODE_DIR (model->priv->root)->cancellable);
 			
+			model_begin_loading (model, model->priv->root);
 			g_file_mount_enclosing_volume (model->priv->root->file, 
 						       G_MOUNT_MOUNT_NONE,
 						       mount_info->operation,
 						       mount_info->cancellable,
 						       (GAsyncReadyCallback)mount_cb,
 						       mount_info);
+			
+			model->priv->mount_info = mount_info;
+			return GEDIT_FILE_BROWSER_STORE_RESULT_MOUNTING;
 		}
 		else
 		{
@@ -2967,6 +2991,14 @@ gedit_file_browser_store_iter_equal (GeditFileBrowserStore * model,
 	return (iter1->user_data == iter2->user_data);
 }
 
+void
+gedit_file_browser_store_cancel_mount_operation (GeditFileBrowserStore *store)
+{
+	g_return_if_fail (GEDIT_IS_FILE_BROWSER_STORE (store));
+	
+	cancel_mount_operation (store);
+}
+
 GeditFileBrowserStoreResult
 gedit_file_browser_store_set_root_and_virtual_root (GeditFileBrowserStore *
 						    model,
@@ -3010,6 +3042,9 @@ gedit_file_browser_store_set_root_and_virtual_root (GeditFileBrowserStore *
 
 		g_object_unref (vfile);
 	}
+	
+	/* make sure to cancel any previous mount operations */
+	cancel_mount_operation (model);
 
 	/* Always clear the model before altering the nodes */
 	model_clear (model, TRUE);
